# -*- coding: utf-8 -*- 
# @Time    : 03/22/2024 
# @Author  : nzooherd 
# @File    : video_service.py 
# @Software: Visual Studio Code

import json
from logging import getLogger
import os
from random import randint
import random
import time
from typing import Tuple

from telethon._impl.client.types import buttons as btns
from nzooherd_pyzen.classes import Singleton

from component.handle_context import HandleContext
from component.video_factory import VideoFactory
from config import Config
from dao.model import VideoTags
from dao.purchase import PurchaseDao
from dao.video_dao import VideoDao
from dao.video_tags_dao import VideoTagsDao
from template import BaseTemplate
from template.no_favorite import NoFavoriteTemplate
from template.search_tag import SearchTagTemplate
from template.video_bought import VideoBoughtTemplate
from template.video_nobuy import VideoNoBuyTemplate
from utills.bytes_util import CallbackBytesRequest, bytes_to_set, int2byte, set_to_bytes

logger = getLogger()

class VideoService(metaclass=Singleton):

    def __init__(self) -> None:
        self.video_factory = VideoFactory()
        self.video_dao = VideoDao()
        self.purchase_dao = PurchaseDao()
        self.video_tags_dao = VideoTagsDao()

        self.handle_context = HandleContext()
    
    async def get_video_message(self, video_md5: str) -> BaseTemplate:
        """
        Return video message
        """
        video = await self.video_dao.get_video_by_md5(video_md5)
        video_tags = await self.video_tags_dao.select_by_video_id(video.id)
        bought = await self.purchase_dao.is_bought(self.handle_context.get_user_id(), video.id)
        
        if not bought:
            video_message = VideoNoBuyTemplate(video.name, [_.tag for _ in video_tags], video.preview_url, video.star)
            btn = btns.Callback("购买")
            btn.request = CallbackBytesRequest("/pc_c", [32])
            btn.request.params = [video.md5.encode("utf-8")]
            buttons = [[btn]]
            logger.info(f"User {self.handle_context.get_user_id()} not bought video {video.md5}")
        else:
            buttons = [[btns.Text(text="/tips"), btns.Text(text="/referral")]]
            video_message = VideoBoughtTemplate(video.name, [_.tag for _ in video_tags], video.preview_url, 
                                                video.online_url, video.magnet_url, video.star)

        video_message.buttons = buttons
        return video_message
    
    async def search_video_init(self, mark_tags: bytes = None) -> Tuple[BaseTemplate, list[list[btns.Button]]]:
        """
        """
        all_tags = set(await self.video_tags_dao.select_all(VideoTags.tag))
        with open(os.path.join(Config.RESOURCE_DIR, "video", "tags.json"), "r") as tags_fp:
            try:
                tag2id = json.load(tags_fp)
            except json.JSONDecodeError:
                tag2id = {}


        id2tag = {v: k for k, v in tag2id.items()}
        if len(tag2id) != len(id2tag):
            raise Exception("Tag id error")

        for tag in all_tags:
            if tag2id.get(tag) is None:
                for i in range(0, 64):
                    if i not in id2tag:
                        tag2id[tag] = i
                        id2tag[i] = tag
                        break
        with open(os.path.join(Config.RESOURCE_DIR, "video", "tags.json"), "w") as tags_fp:
            json.dump(tag2id, tags_fp, indent=4)

        all_id2tag= [(tag2id[tag], tag) for tag in all_tags]
        all_id2tag.sort(key=lambda x: x[0])

        mark_tags = mark_tags or set_to_bytes(set())
        mark_tags_id = bytes_to_set(mark_tags)

        row_tags = 4
        buttons = []
        for i in range(0, len(all_id2tag), row_tags):
            row_buttons = []
            for j in range(i, i + row_tags):
                if j >= len(all_id2tag):
                    break
                tag_id, tag = all_id2tag[j]

                text = tag[0:2]
                if tag_id in mark_tags_id:
                    text += " ✅"

                btn = btns.Callback(text)
                btn.request = CallbackBytesRequest("/tags_c", [1, 8])
                btn.request.params = [int2byte(tag_id), mark_tags]
                row_buttons.append(btn)

            buttons.append(row_buttons) 
        
        btn = btns.Callback("搜索 🔍")
        btn.request = CallbackBytesRequest("/sv_c", [8])
        btn.request.params = [mark_tags]
        buttons.append([btn])
        
        result = SearchTagTemplate()
        result.buttons = buttons
        return result

    async def select_tag(self, mark_tags: bytes, select_tag_id: int) -> BaseTemplate:
        """
        """
        if select_tag_id not in bytes_to_set(mark_tags):
            mark_tags = set_to_bytes(set([select_tag_id]) | bytes_to_set(mark_tags))
        else:
            mark_tags = set_to_bytes(bytes_to_set(mark_tags) - set([select_tag_id]))
        return await self.search_video_init(mark_tags)
       
    async def search_video_scroll(self, mark_tags: bytes, offset: int = 0) -> BaseTemplate:
        """
        """
        with open(os.path.join(Config.RESOURCE_DIR, "video", "tags.json"), "r") as tags_fp:
            tag2id = json.load(tags_fp)

        id2tag = {v: k for k, v in tag2id.items()}
        if len(tag2id) != len(id2tag):
            raise Exception("Tag id error")
        
        mark_tag_ids = bytes_to_set(mark_tags)
        mark_tag_values = [id2tag[tag_id] for tag_id in mark_tag_ids]

        video_ids = await self.video_dao.select_by_tags(mark_tag_values)
        video_ids.sort()
        video = await self.video_dao.select_by_id(video_ids[offset])
        video_tags = await self.video_tags_dao.select_by_video_id(video.id)
        bought = await self.purchase_dao.is_bought(self.handle_context.get_user_id(), video.id)
        if not bought:
            video_message = VideoNoBuyTemplate(video.name, [_.tag for _ in video_tags], video.preview_url, video.star)
            btn = btns.Callback("购买")
            btn.request = CallbackBytesRequest("/pc_c", [32])
            btn.request.params = [video.md5.encode("utf-8")]
            video_message.buttons = [[btn]]
        else:
            video_message = VideoBoughtTemplate(video.name, [_.tag for _ in video_tags], video.preview_url, 
                                    video.online_url, video.magnet_url, video.star)
            video_message.buttons = []

        next_offset = offset + 1
        if next_offset >= len(video_ids):
            next_offset = 0
        last_offset = offset - 1
        if last_offset < 0:
            last_offset = len(video_ids) - 1

        buttons = video_message.buttons
        last_button = btns.Callback("上一个")
        last_button.request = CallbackBytesRequest("/scs_c", [8, 8])
        last_button.request.params = [mark_tags, int2byte(last_offset, 8)]

        next_button = btns.Callback("下一个")
        next_button.request = CallbackBytesRequest("/scs_c", [8, 8])
        next_button.request.params = [mark_tags, int2byte(next_offset, 8)]
        buttons.append([last_button, next_button])

        return video_message

    async def videos_init(self) -> BaseTemplate:
        """
        """
        video_model = await self.video_dao.select_by_offset(0)
        video_tags = await self.video_tags_dao.select_by_video_id(video_model.id)
        video_message = VideoBoughtTemplate(video_model.name, [_.tag for _ in video_tags], video_model.preview_url, 
                                    video_model.online_url, video_model.magnet_url, video_model.star)
        bought = await self.purchase_dao.is_bought(self.handle_context.get_user_id(), video_model.id)
        if not bought:
            video_message = VideoNoBuyTemplate(video_model.name, [_.tag for _ in video_tags], video_model.preview_url, video_model.star)
            btn = btns.Callback("购买")
            btn.request = CallbackBytesRequest("/pc_c", [32])
            btn.request.params = [video_model.md5.encode("utf-8")]
            video_message.buttons = [[btn]]
        else:
            video_message = VideoBoughtTemplate(video_model.name, [_.tag for _ in video_tags], video_model.preview_url, 
                                    video_model.online_url, video_model.magnet_url, video_model.star)
            video_message.buttons = []
        
        next_button = btns.Callback("下一个")
        next_button.request = CallbackBytesRequest("/vo_c", [8, 8])
        next_button.request.params = [int2byte(video_model.id, 8), int2byte(1, 8)]
        video_message.buttons.append([next_button])

        return video_message
    
    async def videos_scroll(self, video_id: int, scorll_dir: int) -> BaseTemplate:
        """
        """
        logger.error(f"video_id: {video_id}, scorll_dir: {scorll_dir}")
        video_model = await self.video_dao.select_video_scorll(video_id, scorll_dir)
        video_tags = await self.video_tags_dao.select_by_video_id(video_model.id)
        bought = await self.purchase_dao.is_bought(self.handle_context.get_user_id(), video_model.id)
        if not bought:
            video_message = VideoNoBuyTemplate(video_model.name, [_.tag for _ in video_tags], video_model.preview_url, video_model.star)
            btn = btns.Callback("购买")
            btn.request = CallbackBytesRequest("/pc_c", [32])
            btn.request.params = [video_model.md5.encode("utf-8")]
            video_message.buttons = [[btn]]
        else:
            video_message = VideoBoughtTemplate(video_model.name, [_.tag for _ in video_tags], video_model.preview_url, 
                                    video_model.online_url, video_model.magnet_url, video_model.star)
            video_message.buttons = []
        
        buttons = []
        if (scorll_dir == 1 or await self.video_dao.select_video_scorll(video_model.id, 1 - scorll_dir)):
            last_button = btns.Callback("上一个")
            last_button.request = CallbackBytesRequest("/vo_c", [8, 8])
            last_button.request.params = [int2byte(video_model.id, 8), int2byte(0, 8)]
            buttons.append(last_button)

        if (scorll_dir == 0 or await self.video_dao.select_video_scorll(video_model.id, 1 - scorll_dir)):
            next_button = btns.Callback("下一个")
            next_button.request = CallbackBytesRequest("/vo_c", [8, 8])
            next_button.request.params = [int2byte(video_model.id, 8), int2byte(1, 8)]
            buttons.append(next_button)

        video_message.buttons.append(buttons)

        return video_message
 

    async def favoriate_video_init(self) -> BaseTemplate:
        """
        """
        random_seed = int(time.time() * 1000000)
        random.seed(random_seed)

        video_ids = await self.video_dao.select_has_bought(uid=self.handle_context.get_user_id())
        if (video_ids is None) or (len(video_ids) < 1):
            return NoFavoriteTemplate()


        random.shuffle(video_ids)

        #TODO(Use redis cache if performance is not good)
        video = await self.video_dao.select_by_id(video_ids[0])
        video_tags = await self.video_tags_dao.select_by_video_id(video.id)
        video_message = VideoBoughtTemplate(video.name, [_.tag for _ in video_tags], video.preview_url, 
                                    video.online_url, video.magnet_url, video.star)
        buttons = []
        if len(video_ids) != 1:
            last_button = btns.Callback("上一个")
            last_button.request = CallbackBytesRequest("/fr_c", [16, 8])
            last_button.request.params = [str(random_seed).encode("utf-8"), int2byte(len(video_ids) - 1, 8)]

            next_button = btns.Callback("下一个")
            next_button.request = CallbackBytesRequest("/fr_c", [16, 8])
            next_button.request.params = [str(random_seed).encode("utf-8"), int2byte(1, 8)]

            buttons.append([last_button, next_button])

        video_message.buttons = buttons
        return video_message
 

    async def favoriate_video_scroll(self, randomseed: int, offset: int) -> BaseTemplate:
        """
        """
        random.seed(randomseed)
        video_ids = await self.video_dao.select_has_bought(uid=self.handle_context.get_user_id())
        random.shuffle(video_ids)


        video = await self.video_dao.select_by_id(video_ids[offset])
        video_tags = await self.video_tags_dao.select_by_video_id(video.id)
        video_message = VideoBoughtTemplate(video.name, [_.tag for _ in video_tags], video.preview_url, 
                                    video.online_url, video.magnet_url, video.star)
        
        next_offset = offset + 1
        if next_offset >= len(video_ids):
            next_offset = 0
        last_offset = offset - 1
        if last_offset < 0:
            last_offset = len(video_ids) - 1

        buttons = []
        last_button = btns.Callback("上一个")
        last_button.request = CallbackBytesRequest("/fr_c", [16, 8])
        last_button.request.params = [str(randomseed).encode("utf-8"), int2byte(last_offset, 8)]

        next_button = btns.Callback("下一个")
        next_button.request = CallbackBytesRequest("/fr_c", [16, 8])
        next_button.request.params = [str(randomseed).encode("utf-8"), int2byte(next_offset, 8)]
        buttons.append([last_button, next_button])

        video_message.buttons = buttons
        return video_message

    async def random_video(self) -> Tuple[BaseTemplate, list[list[btns.Button]]]:
        """
        Return random video
        """
        total_count = await self.video_dao.activate_count()
        if total_count < 1:
            raise Exception("Video count error")

        random_video = await self.video_dao.select_by_offset(randint(0, total_count - 1))
        return await self.get_video_message(random_video.md5)